#
#
#
module Postfix::Rate::Policyd

  class DataBackend
    include Celluloid::Internals::Logger
    include Heuristics
    def debug(string)
      Celluloid.logger.debug(string) if Celluloid.logger
    end

    class EntityFactory
      include Celluloid


      def fetch_by_identity(identity)
        Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: identity)
      end

      def fetch_by_address(address)
        Postfix::Rate::Models::ClientAddress.find_or_create(address: address)
      end
    end

    #
    #
    #
    class Response
      attr_reader :message

      def initialize(ack, message)
        @ack, @message = ack, message
      end

      def nack?
        !ack?
      end

      def ack?
        @ack
      end
    end

    #
    #
    #
    def initialize
    end
    
    def entity_factory
      @entity_supervisor ||= EntityFactory.supervise as: :entity_factory
      #entity_supervisor.actors.first
      Celluloid::Actor[:entity_factory]
    end


    def fetch_entity(request, refresh = false)
      if request.entity
        request.entity.refresh if refresh
      else
        if request.authenticated?
          entity = entity_factory.future.fetch_by_identity request.authenticated_identity
        else
          entity = entity_factory.future.fetch_by_address request.client_address
        end
        request.entity = entity.value
      end
      request.entity
    end


    def normalized_tracker(value, request, increment = 1)
      tracker, policy = request.entity.tracker, request.entity.policy

      ratio = quota_ratio(tracker.send(value), policy.interval, tracker.last_update)
      debug "#{request.label}: check {#{value}: #{ratio}(+#{increment})/#{policy.send(value)}, interval #{humanize_interval(policy.interval)}}"
      (ratio + increment)
    end

    def update_tracker(value, request, increment = 1)
      request.entity.tracker.set(value =>  normalized_tracker(value, request, increment))
    end

    def tracker_exceeded?(value, request, increment = 1)
      (normalized_tracker(value, request, increment) > request.entity.policy.send(value)) ? true : false
    end

    def check_blacklisting(request)
      fetch_entity_address(request) unless request.entity_address
      return fail(request, "The IP address '#{request.client_address} has been blacklisted.'")  if address_check_blacklist(request)
    end

    def check_disabled_identity(request)
      return fail(request, "Your email account was disabled on #{request.entity.disable_date}: #{request.entity.disable_reason}") if
          request.entity.disabled
    end

    def check_initial_quota(request)

      entity = fetch_entity request
      failed = check_disabled_identity request
      return failed if failed&.nack?

      failed = check_blacklisting request
      return failed if failed&.nack?

      if entity.policy.count_tracker?
        if tracker_exceeded?(:count, request)
          return fail request, "exceeded count quota #{entity.tracker.count}(+1)/#{entity.policy.count} messages, interval #{humanize_interval(entity.policy.interval)}"
        end
      end

      success
    end

    def check_rcpt_quota(request)
      entity = fetch_entity request

      if entity.policy.recipients_tracker?
        if tracker_exceeded?(:recipients, request, request.recipient_count)
          return fail request, "exceeded recipients quota #{entity.tracker.recipients}(+#{request.recipient_count})/#{entity.policy.recipients}, interval #{humanize_interval(entity.policy.interval)}"
        end
      end

      success
    end

    def check_final_quota_and_update(request)

      entity = fetch_entity(request)

      if entity.policy.volume_tracker?
        if tracker_exceeded?(:volume, request, request.size)
          return fail request, "exceeded volume quota #{entity.tracker.volume}(+#{request.size})/#{entity.policy.volume} bytes, interval #{humanize_interval(entity.policy.interval)}"
        end
      end

      update_quota(request)
      success
    end

    def humanize_interval secs
      [[60, :seconds], [60, :minutes], [24, :hours], [1000, :days]].map{ |count, name|
        if secs > 0
          secs, n = secs.divmod(count)
          "#{n.to_i} #{name}" if n > 0
        end
      }.compact.reverse.join(' ')
    end

    def quota_details(tracker, policy)
      details = []
      details << "count: #{tracker.count.round(2)}/#{policy.count_tracker? ? policy.count : '-'}"
      details << "recipients: #{tracker.recipients.round(2)}/#{policy.recipients_tracker? ? policy.recipients : '-'}"
      details << "volume: #{tracker.volume.round}/#{policy.volume_tracker? ? policy.volume : '-'}"
      details << "interval: #{humanize_interval(policy.interval)}"
      details.join(', ')
    end

    def update_quota(request)

      entity = request.entity

      Sequel::Model.db.transaction do

        entity.tracker.lock!
        update_tracker(:count, request)
        update_tracker(:recipients, request, request.recipient_count)
        update_tracker(:volume, request, request.size)

        check_heuristics(request) if request.authenticated? && respond_to?(:check_heuristics)

        info "#{request.label}: update -> (#{quota_details(entity.tracker, entity.policy)})"

        entity.tracker.last_update = Time.now
        entity.tracker.save_changes
      end
    end

    private

    def self.quota_ratio(current_value, interval, last_seen)

      current_value ||= 0
      last = last_seen.is_a?(Time) ? last_seen : (last_seen || DateTime.new(2013, 1, 1)).to_time

      elapsed = (Time.now - last).to_i
      (elapsed = 0) if elapsed < 0

      if elapsed > interval
        return 0
      end
      current_value.to_f * (1.00 - (elapsed.to_f / interval.to_f)).round(3)
    end

    def quota_ratio(current_value, interval, last_seen)
      self.class.quota_ratio(current_value, interval, last_seen)
    end

    def fail(request, message)
      warn "#{request.label}: #{message}"
      Response.new(false, message)
    end

    def success
      Response.new(true, nil)
    end

    def fetch_entity_address(request)
      request.entity_address =
        Postfix::Rate::Models::NetworkAddress.find(address: request.client_address)
    end

  end
end
