module Postfix::Rate::Policyd
  module Heuristics
    BLACKLIST_PERSISTENCE = 7 * 86400

    extend Concern

    included do

    end

    class Store


      attr_reader :data
      attr_reader :request
      attr_reader :entity

      def initialize(request)
        data = request.entity.tracker.heuristics

        @request = request
        @entity = request.entity
        @data = data || new_data_for(request)
      end

      def addresses
        data[:addresses]
      end

      def add_address(address)
        addresses[address] = {first_seen: Time.now, last_seen: Time.now, total_count: 0, last_count: 0}
      end

      def has_address?(address)
        addresses.has_key?(address)
      end

      def addresses_count
        addresses.count
      end

      def update_address(address)
        a = addresses[address]
        a[:total_count] += 1
        a[:last_count] = DataBackend.quota_ratio(a[:last_count], entity.policy.h_interval, a[:last_seen]) + 1
        a[:last_seen] = Time.now
      end


      def check_activity_for(address)
        # first and only address sending mail for this user
        return false if addresses_count == 1

        a = addresses[address]

        # grace addresses we have known for over period of time
        return false if (Time.now - a[:first_seen]) >  entity.policy.h_grace_threshold

        return "The address '#{address}' has sent over #{entity.policy.h_interval_count} messages in the last #{entity.policy.h_interval} seconds" if
            a[:last_count] && (a[:last_count] > entity.policy.h_interval_count)
        false
      end

      def check_address(address)

        if has_address?(address)
          update_address(address)
          check_activity_for(address)
        else
          if addresses_count >= entity.policy.h_max_addresses

            if expire_old_address
              add_address address
            else
              return "Too many addresses are active in the last #{entity.policy.h_activity_threshold} seconds"
            end
          else
            add_address address
          end
          update_address(address)
          check_activity_for(address)
        end
      end

      def oldest_seen_address
        addresses.keys.inject(addresses.keys.first) {|current, add| (addresses[add][:last_seen] < addresses[current][:last_seen]) ? add : current }
      end

      def expire_old_address
        oldest = addresses[oldest_seen_address]

        if((Time.now - oldest[:last_seen]) > entity.policy.h_activity_threshold)
          addresses.delete(oldest_seen_address)
          true
        else
          false
        end
      end

      private


      def new_data_for(request)

        @data = entity.tracker.heuristics = {
            addresses: {}
        }
      end
    end

    #
    #
    #
    def check_heuristics(request)

      h = Store.new(request)
      message = h.check_address(request.client_address)
      if message && request.entity.policy.heuristics && !address_check_whitelist(request)
        blacklist_address!(request.client_address, message)
        warn "Heuristics: identity '#{request.entity.label}' has been disabled: #{message}"
        request.entity.disable!(message)
      else
        request.entity.tracker.heuristics = h.data
      end
    end


    def blacklist_address!(address, reason)
      warn "Heuristics: address '#{address}' has been blacklisted: #{reason}"
      Postfix::Rate::Models::NetworkAddress.create(address: address, notes: reason, inserted: Time.now)
    end

    def address_check_whitelist(request)
      if (request.entity_address && request.entity_address.whitelisted?)
        info "Heuristics: address #{request.client_address} is whitelisted."
        true
      else
        false
      end
    end

    def address_check_blacklist(request)
      if (request.entity_address && request.entity_address.blacklisted?)
        if(Time.now -  request.entity_address.inserted) > BLACKLIST_PERSISTENCE
          request.entity_address.destroy
          false
        else
          request.entity_address.notes
        end
      end
    end

  end
end