
module Postfix::Rate::Models

  def self.init(config = nil)
    create_policies_schema
    create_trackers_schema
    create_authenticated_identities_schema
    create_client_addresses_schema
    create_network_addresses_schema

    require "postfix/rate/models"

    initialize_policies(config)
    initialize_trackers(config)
  end

  def self.create_authenticated_identities_schema
    unless Sequel::Model.db.table_exists?(:authenticated_identities)
      Sequel::Model.db.create_table :authenticated_identities do
        primary_key :id
        String :identity, size: 128, unique: true

        String  :notes, null: true
        boolean :disabled, null: false, default: false
        boolean :static, null: false, default: false
        String  :disable_reason, null: true
        DateTime    :disable_date,   null: true
        foreign_key :policy_id,  :policies, null: false
        foreign_key :tracker_id, :trackers, null: true
      end
      true
    end
  end

  def self.create_client_addresses_schema
    unless Sequel::Model.db.table_exists?(:client_addresses)
      Sequel::Model.db.create_table :client_addresses do
        primary_key :id
        String :address, size: 64, unique: true

        String      :notes, null: true
        boolean     :disabled, null: false, default: false
        boolean     :static, null: false, default: false
        String      :disable_reason, null: true
        DateTime    :disable_date,   null: true
        foreign_key :policy_id,  :policies, null: false
        foreign_key :tracker_id, :trackers, null: true
      end
      true
    end
  end

  def self.create_network_addresses_schema
    unless Sequel::Model.db.table_exists?(:network_addresses)
      Sequel::Model.db.create_table :network_addresses do
        primary_key :id
        String :address, size: 64, unique: true

        String :flags, fixed: true, size: 2, null: false, default: 'b'
        String :notes, null: true
        DateTime :inserted,   null: true
      end
      true
    end
  end

  def self.create_policies_schema
    unless Sequel::Model.db.table_exists?(:policies)
      Sequel::Model.db.create_table :policies do
        primary_key :id
        String  :name, size: 64, unique: true, null: false
        String  :description
        Fixnum  :interval, null: false, default: 86400

        boolean :heuristics, null: false, default: true

        # limits for, null to disable this quota
        Fixnum  :count, null: true
        Fixnum  :recipients, null: true
        Bignum  :volume, null: true


        # how many different IP addresses are we going to keep
        # when this many different _active_ IP address are seen, the user is temporarily disabled
        Fixnum  :h_max_addresses, null: false, default: 5

        # an address is considered active if it has sent mail in the last h_activity_threshold seconds
        Fixnum  :h_activity_threshold, null: false, default: 4*3600

        # if an address sends more than h_interval_count messages in h_interval, gets included into a blacklist
        # of blocked addresses
        Fixnum  :h_interval, null: false, default: 1800
        Fixnum  :h_interval_count, null: false, default: 50

        # if we have known this address longer than h_grace_threshold, we trust it and do not blacklist
        Fixnum  :h_grace_threshold, null: false, default: 15*86400

      end
      true
    end
  end

  def self.create_trackers_schema
    unless Sequel::Model.db.table_exists?(:trackers)
      Sequel::Model.db.create_table :trackers do

        primary_key :id

        Float   :count, null: true
        Float   :recipients, null: true
        Bignum   :volume, null: true
        DateTime :last_update, null: false
        DateTime :first_seen,  null: false
        String   :heuristics, text: true, null: true
      end
      true
    end
  end

  def self.initialize_policies(config = nil)
    if config && config.policies.any?
      config.policies.each do | pol |
        policy = Policy.find(name: pol['name'])
        policy ||= Policy.new
        pol.each do | k, v |
          policy.send("#{k}=", v)
        end
        policy.save
      end
    end
  end

  def self.initialize_trackers(config = nil)
    if config && config.trackers
      if config.trackers['identities']
        config.trackers['identities'].each do | id |
          authid = AuthenticatedIdentity.find(identity: id['identity'])
          authid ||= AuthenticatedIdentity.new
          authid.identity = id['identity']
          authid.notes = id['notes']
          authid.policy = Policy.find(name: id['policy'])
          raise StandardError.new("Policy #{id['policy']} not found") unless authid.policy
          authid.save
        end
      end

      if config.trackers['addresses']
        config.trackers['addresses'].each do | add |
          cladd = ClientAddress.find(address: add['address'])
          cladd ||= ClientAddress.new
          cladd.address = add['address']
          cladd.notes = add['notes']
          cladd.policy = Policy.find(name: add['policy'])
          raise StandardError.new("Policy #{add['policy']} not found") unless cladd.policy
          cladd.save
        end
      end
    end
  end
end